declare name "Advanced Crossfeed";
declare id "adv_xfeed";
declare author "Radoslaw Szkodzinski";
declare email "astralstorm@gmail.com";
declare version "1.8";

import("filter.lib");
import("math.lib");

// LICENCE: ISC (like BSD 2-clause, but shorter.)
/*
Copyright (c) 2013, Radosław Szkodziński <astralstorm@gmail.com>

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee
is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE
FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

Delaylen = 512; // enough for 0.75 ms delay @ 192000 HZ. Must be > 2.0 Order
delayfun = fdelay5;

// Defaults are tweaked by ear.
freq = nentry("Corner frequency (Hz)[unit:Hz]", 1500, 150, 6000, 10);
lgain = nentry("LF feed level (dB)[unit:dB]", -3, -48, 0, 0.1) : db2linear;
hgain = nentry("HF feed level (dB)[unit:dB]", -9, -48, 0, 0.1) : db2linear;
gain = nentry("Gain (dB)[unit:dB]", -3.0, -24, 12, 0.1) : db2linear;
itd_delay = nentry("Inter-aural time delay (us)[unit:us]", 33.5, 0, 750, 1) : *(0.000001);
alpha = nentry("Filter steepness", 0.75, 0.01, 6, 0.01);
center_focus = nentry("Center control (percent)[unit:%%]", 33.5, 0, 100, 0.1) : *(0.01);
curvature = nentry("Directionality (percent)[unit:%%]", 15, 0, 100, 1) : *(0.01);

// Knee frequency for the filter. Only approximate, as is Kaiser vs a bessel eq.
ft = freq/SR;

// Pattern match doesn't work here, so a slow if was necessary.
// Trimmed to prevent clipping, which could happen at some knee frequencies.
sinc(M, i) = if(i == M/2, 2*ft, sin(2*PI*ft*(i - M/2))/(PI*(i - M/2))) : min(2*ft);
sinc_inv(M, i) = if(i == M/2, 1-2*ft, 0-sin(2*PI*ft*(i - M/2))/(PI*(i - M/2))) : min(1-2*ft);

fact(0) = 1;
fact(1) = 1;
fact(k) = k * fact(k-1);

// Modified Bessel function of the first kind, approximated.
I0(K, x) = I0_part(K, K, x)
with {
    I0_part(K, 0, x) = 1;
    I0_part(K, i, x) = x^(2*i)/(2^(2*1)*fact(i)^2) + I0_part(K, i-1, x);
};

// Without the float twiddle this will explode.
// Trimmed to prevent clipping, which did happen at some knee frequencies.
kaiser(N, alpha, N+i) = 0;
kaiser(N, alpha, i) = I0(15, PI*alpha*sqrt(1 - (2*i/float(N))^2))/I0(15, PI*alpha) : min(1.0);

kaiser2(N, alpha, i) = kaiser(N, alpha, i - N/2);

// Must be even for the highpass to work properly.
// If you increase it, make sure to increase Delaylen too to the next power of 2.
// Directly affects CPU usage. If too low, filter becomes allpass non-complementary.
// Transition band might have artifacts in such case and filter becomes inaccurate at lower corner frequencies.
// If it's very high, the faust compile will take a very long time.
Order = 180;

lp_fir_coeffs = par(i,Order,sinc(Order, i) * kaiser2(Order, alpha, i));
hp_fir_coeffs = par(i,Order,sinc_inv(Order, i) * kaiser2(Order, alpha, i));

//mid(x,y) = (x+y)/2.0;
//side(x,y) = (x-y)/2.0;
// Parabolic weighting function, also rejects side cancellation
mid(x,y) = pow(abs(x-y)/2.0, curvature)*((x+y)/2.0) + (1.0-pow(abs(x-y)/2.0, curvature))*x;
side(x,y) = ((x-y)/2.0)*pow(abs(x+y)/2.0, curvature) + (1.0-pow(abs(x+y)/2.0, curvature))*x;

// fdelay5 can be replaced with a faster fdelay, obviously after fixing the FIR delay correction.
// This will very slightly reduce sound quality and bring a minor speedup.
channel_process = _ <: fir(lp_fir_coeffs), fir(hp_fir_coeffs) : *(lgain),*(hgain) : + : delayfun(Delaylen, itd_delay * SR);
// Delay correction to avert FIR combing.
crossfeed(x,y) = (
    (channel_process(side(y,x)) + delayfun(Delaylen, Order/2, x*(1.0-center_focus) + mid(x,y)*center_focus)),
    (channel_process(side(x,y)) + delayfun(Delaylen, Order/2, y*(1.0-center_focus) + mid(x,y)*center_focus))
);

// 100% correlated white noise for testing.
//process = noise <: fir(lp_fir_coeffs) + fir(hp_fir_coeffs); // FIR allpass test. Too low filter order fails it.
//process = noise <: crossfeed;
//process(x,y) = crossfeed(x,y) : (*(gain),*(gain));
process(x,y) = (x*(gain),y*(gain)) : crossfeed;


